#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <string.h>

#define MAXBUFSIZE 100
#define MAXCONN 3

void sig_chld(int signo)
{
  pid_t pid;
  int stat;
  while ((pid = waitpid(-1, &stat, WNOHANG)) > 0) 
    printf("child %d terminated.\n", pid);
  return ;
}



int process_request(int fd)
{
  char buf[MAXBUFSIZE];
  int num;

  if((num = read(fd, buf, MAXBUFSIZE)) > 0) 
    write(fd, buf, num);
  else
    return 0;

  return 1;
}


void AcceptNewClient(int sockfd, int *client_array, int Max_Connections, fd_set *allset, int *maxfd)
{
  int rec_sock, i;
  
  if( (rec_sock = accept(sockfd, NULL, NULL)) < 0)
    if (errno == EINTR)
      return;
    else 
    {
      perror("accept");
      exit(1);
    }


  for (i=0; i<Max_Connections; i++) 
    if (client_array[i] < 0) 
    {
      client_array[i] = rec_sock; 
      FD_SET(client_array[i], allset);
      break;
    }

  if (i== Max_Connections) 
  {
    fprintf(stderr, "Too many connections.\n");
    close(rec_sock);
  }
  else if (rec_sock > *maxfd) 
    *maxfd = rec_sock;

  return;
}


int main(int argc, char **argv)
{
  int sockfd, client[MAXCONN], i, maxfd;
  socklen_t len;
  struct sockaddr_in addr;
  struct sigaction sa;
  char address[INET_ADDRSTRLEN];
  fd_set allset, rset; 

  if (argc < 2) 
  {
    fprintf(stderr,"USAGE: %s port.\n", argv[0]);
    exit(-1);
  }

  /* Handle SIGCHLD */
  sa.sa_handler = sig_chld;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(SIGCHLD, &sa, NULL);


  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
  {
    perror("Cannot get socket");
    exit(1);
  }

  memset(&addr, 0, sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons((short)atoi(argv[1]));

  if (bind(sockfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) 
  {
    perror("bind");
    exit(1);
  }

  
  len = sizeof(addr);
  if (getsockname(sockfd, (struct sockaddr *)&addr, &len) < 0)
  {
    perror("can't get name");
    exit(1);
  }
  
  printf("ip = %s, port = %hd\n", inet_ntop(AF_INET, &addr.sin_addr, address, sizeof(address)), ntohs(addr.sin_port));

  if (listen(sockfd, 5) < 0) 
  {
    perror("listen");
    exit(1);
  }
 
  for (i=0; i<MAXCONN; i++) 
    client[i] = -1;

  FD_ZERO(&allset);
  FD_SET(sockfd, &allset);
  maxfd = sockfd;

  for(;;) 
  {
    rset = allset;
    select(maxfd+1, &rset, NULL, NULL, NULL);
    if (FD_ISSET(sockfd, &rset)) /* somebody tries to connect */
      AcceptNewClient(sockfd, client, MAXCONN, &allset, &maxfd);

    for (i=0; i<MAXCONN; i++) 
      if (client[i] >= 0 && FD_ISSET(client[i], &rset) && !process_request(client[i])) 
      {/* Client exits */
	close(client[i]);
	FD_CLR(client[i], &allset);
	client[i] = -1;
      }
  }

}




